@@ -3,6 +3,7 @@ from __future__ import annotations
 import asyncio
 import logging
 import json
+import random
 from typing import (
     Optional, Union,
     Tuple,
@@ -11,6 +12,7 @@ from typing import (
 
 import aiohttp
 from aiohttp import web
+from aiohttp_session import get_session, STORAGE_KEY
 
 from ai.backend.client.exceptions import BackendAPIError, BackendClientError
 from ai.backend.client.request import Request
@@ -128,7 +130,7 @@ async def web_handler(request, *, is_anonymous=False) -> web.StreamResponse:
             # We treat all requests and responses as streaming universally
             # to be a transparent proxy.
             api_rqst = Request(
-                api_session, request.method, path, request.content,
+                request.method, path, request.content,
                 params=request.query,
                 override_api_version=request_api_version)
             if 'Content-Type' in request.headers:
@@ -196,7 +198,7 @@ async def web_plugin_handler(request, *, is_anonymous=False) -> web.StreamRespon
                 content = json.dumps(body).encode('utf8')
             request_api_version = request.headers.get('X-BackendAI-Version', None)
             api_rqst = Request(
-                api_session, request.method, path, content,
+                request.method, path, content,
                 params=request.query,
                 content_type=request.content_type,
                 override_api_version=request_api_version)
@@ -238,16 +240,34 @@ async def web_plugin_handler(request, *, is_anonymous=False) -> web.StreamRespon
 
 async def websocket_handler(request, *, is_anonymous=False) -> web.StreamResponse:
     path = request.match_info['path']
+    session = await get_session(request)
+    app = request.query.get('app')
+
+    # Choose a specific Manager endpoint for persistent web app connection.
+    api_endpoint = None
+    should_save_session = False
+    _endpoints = request.app['config']['api']['endpoint'].split(',')
+    _endpoints = [e.strip() for e in _endpoints]
+    if session.get('api_endpoints', {}).get(app):
+        if session['api_endpoints'][app] in _endpoints:
+            api_endpoint = session['api_endpoints'][app]
+    if api_endpoint is None:
+        api_endpoint = random.choice(_endpoints)
+        if 'api_endpoints' not in session:
+            session['api_endpoints'] = {}
+        session['api_endpoints'][app] = api_endpoint
+        should_save_session = True
+
     if is_anonymous:
-        api_session = await asyncio.shield(get_anonymous_session(request))
+        api_session = await asyncio.shield(get_anonymous_session(request, api_endpoint))
     else:
-        api_session = await asyncio.shield(get_api_session(request))
+        api_session = await asyncio.shield(get_api_session(request, api_endpoint))
     try:
         async with api_session:
             request_api_version = request.headers.get('X-BackendAI-Version', None)
             params = request.query if request.query else None
             api_rqst = Request(
-                api_session, request.method, path, request.content,
+                request.method, path, request.content,
                 params=params,
                 content_type=request.content_type,
                 override_api_version=request_api_version)
@@ -256,6 +276,9 @@ async def websocket_handler(request, *, is_anonymous=False) -> web.StreamRespons
                 await down_conn.prepare(request)
                 web_socket_proxy = WebSocketProxy(up_conn.raw_websocket, down_conn)
                 await web_socket_proxy.proxy()
+                if should_save_session:
+                    storage = request.get(STORAGE_KEY)
+                    await storage.save_session(request, down_conn, session)
                 return down_conn
     except asyncio.CancelledError:
         raise
